package CatalystX::state51::ControllerRole::DataGridHandler;
use MooseX::MethodAttributes::Role;
use MooseX::Types::Moose qw/HashRef ArrayRef/;
use MooseX::Types::Common::String qw/ NonEmptySimpleStr /;
use MooseX::Types::Common::Numeric qw/PositiveInt/;
use MooseX::Types::LoadableClass qw/ LoadableClass /;
use Moose::Autobox;
use Data::Dumper;
use Try::Tiny;
use Data::Pageset::Render;
use Carp qw/ confess /;
use MIME::Base64;
use Storable 'nfreeze';
use Text::SimpleTable;
use namespace::autoclean;

#with qw/
#    CiSupport::ControllerRole::FlashMessages
#/;

has s51_class => (
    isa => LoadableClass,
    coerce => 1,
    required => 1,
    handles => {
        fetch_object => 'fetch',
        fetch_object_ro => 'fetch_readonly',
        load_object_by_id => 'load_oids',
        get_object_pager => 'fetch_pager',
        new_order => 'new',
        type_human => 'type_human',
    }
);

has filter_key => (
    is            => 'rw',
    isa           => NonEmptySimpleStr,
    required      => 1,
);

has filter_form_class => (
    isa => LoadableClass,
    coerce => 1,
    required => 1,
    handles => {
        filter_form_instance => 'new',
    }

);

has filter_instance_params_from_stash => (
    is => 'ro',
    isa => ArrayRef,
    default => sub { return [] },
);

sub get_filter_form_instance {
    my ($self, $c, @args) = @_;
    my $args = Moose::Object->BUILDARGS(@args);
    foreach ($self->filter_instance_params_from_stash->flatten){
        $args->{$_} = $c->stash->{$_};
    }
    my $reset_url = '/' . $c->req->path . '/reset_filter';
    $args->{reset_url} = "$reset_url"; # stringify
    return $self->filter_form_instance($args);
};


has default_fetch_params => (
    isa => HashRef,
    default => sub { {} },
    traits => ['Hash'],
    handles => {
        default_fetch_params => 'elements'
    },
);

has default_items_per_page => (
    isa => PositiveInt,
    is => 'ro',
    default => 50,
);

sub set_filter_params {
    my ($self, $c, $form) = @_;
    confess('No $c given') unless $c;
    $form ||= $self->get_filter_form_instance($c);
    $form->process( $c->req->parameters );

    $c->stash(filter_form => $form, fillinform => 1);
    my $fif = $form->fif;

    if ($form->validated) {
        $c->log->debug("\tForm validated ok");
        $self->store_filter_params($c, $fif);
        return 1;
    }

    if ($self->can('set_flash')) {
        $self->set_flash($c,'error','Filter form data is not valid.');
    }

    foreach my $field ($form->fields) {
        if (scalar @{$field->errors}) {
            $c->log->debug("\tError: Deleting field " . $field->name
                . " from params");
            delete $c->req->parameters->{$field->name};
            delete $fif->{$field->name};
        }
    }

    $self->store_filter_params($c, $fif);

    return;
}

sub reset_filter_params {
    my ($self, $c) = @_;
    confess('No $c given') unless $c;
    delete($c->session->{filter_params}->{$self->filter_key});
}

sub get_filter_form {
    my ($self, $c) = @_;
    my $form = $self->get_filter_form_instance($c);
    $form->setup_form($self->get_filter_params($c));
    return $form;
}

sub get_filter_params {
    my ($self, $c) = @_;
    confess('No $c given') unless $c;
    my $filter_params = $c->session->{filter_params};
    return $filter_params->{$self->filter_key}
        if (defined $filter_params and defined $filter_params->{$self->filter_key} );
    return {};
}

sub store_filter_params {
    my ($self, $c, $params) = @_;
    my $filter_params = $c->session->{filter_params} ||= {};

    # Don't explode the session: our catalyst_sessions MySQL value field is
    # type TEXT, so can only hold ~65535 bytes of frozen base64_encoded data.
    # Here we cut unreasonable stuff down to size and inform the user instead
    # of blorking the session and logging out.

    my $max = 65000; # Leave a few hundred bytes for other session contents

    if ( length(encode_base64( nfreeze($params) )) > $max ) {
        $c->log->debug(
            "Stored form params would be longer than $max bytes, losing some"
        );
        if($self->can('set_flash')) {
            $self->set_flash($c, 'error',
                "Your form input was too lengthy;"
                . " some paramters will have been discarded" );
        }

        my @fields_by_size =
            sort { length(encode_base64(nfreeze([$params->{$b}])))
                <=> length(encode_base64(nfreeze([$params->{$a}]))) }
            keys %$params;

        do {
            my $fld = shift @fields_by_size;
            $c->log->debug("Dropping field '$fld'");
            delete $params->{$fld};
        } while ( length(encode_base64( nfreeze($params) )) > $max );
    }

    return 0 unless keys %$params;

    $filter_params->{$self->filter_key} = $params;
}

# left is a default
# right is an override
sub _merge {
    my ($left, $right) = @_;
    my %merged;
    foreach my $k (sort keys(%$left), keys(%$right)) {
        $merged{$k} = (exists $right->{$k} ? $right->{$k} : $left->{$k}) || undef;
    }
    return \%merged;
}

sub get_fetch_params {
    my ($self, $c) = @_;
    confess('No $c given') unless $c;
    my %merged = %{ _merge( { $self->default_fetch_params } , $self->get_filter_params($c)) };

    foreach (keys %merged) {
        if((not length $merged{$_}) || $merged{$_} eq 'ignore') {
            delete $merged{$_};
        }
        elsif ($merged{$_} eq 'yes') {
            $merged{$_} = 1
        }
        elsif ($merged{$_} eq 'no') {
            $merged{$_} = 0;
        }
    }
    if ($c->stash->{fetch_params_hook}) {
        %merged = %{ _merge(\%merged, $c->stash->{fetch_params_hook}) };
    }
    return \%merged;
}

sub get_fetch_params_sorted {
    my ($self, $c) = @_;
    confess('No $c given') unless $c;
    my $fetch_params = $self->get_fetch_params($c);
    my $order_by = $c->req->param('order_by')|| $fetch_params->{order_by} || 'id';
    my $desc = 1;
    $desc = $fetch_params->{desc} if (defined $fetch_params->{desc});
    $desc = (defined $c->req->param('desc'))? $c->req->param('desc') : $desc;
    $fetch_params->{order_by} = $order_by;
    $fetch_params->{desc} = $desc;
    return $fetch_params;
}

sub get_pager_params {
    my ($self, $c) = @_;
    confess('No $c given') unless $c;
    my $fetch_params = $self->get_fetch_params_sorted($c);
    my $page = $c->req->param('page')||1;
    my $items_per_page = $c->req->param('items_per_page')|| $self->default_items_per_page;
    $c->stash(items_per_page => $items_per_page);
    $c->stash(page => $page);
    my $uri = $c->req->uri_with({
        page => '%p',
        order_by => $fetch_params->{order_by},
        desc => $fetch_params->{desc},
        items_per_page => $c->stash->{items_per_page}
    });
    $uri =~ s/%25/%/g;
    return {
        current_page     => $page,
        entries_per_page => $items_per_page,
        link_format      => "&nbsp;<a href=\"$uri\">%a</a>&nbsp;",
        current_page_format => '[%a]',
    };
}

sub base_grid : Chained('base') PathPart('') CaptureArgs(0) {
    my ($self, $c) = @_;
}

sub default : Chained('base_grid') PathPart('') Args(0) {
    my ( $self, $c ) = @_;
    $c->forward('default_grid');
    if ($c->req->param('csv')) {
        $c->stash(csv => { data => $c->stash->{pager}->for_csv }, current_view => 'CSV');
    }
}

sub default_grid : Action {
    my ( $self, $c ) = @_;
    my $pager = $self->get_pager_instance($c);
    $pager->objects; #pre-fetch objects and catch issues

    my $fetch_params = $self->get_fetch_params_sorted($c); # doing this again as the fetch deletes stuff from the HashRef
    $c->stash(
        data_grid_params => $fetch_params,
        pager => $pager,
    );

    $c->stash(filter_form => $self->get_filter_form($c) )
        unless $c->stash->{filter_form};
}

sub get_pager_instance {
    my ( $self, $c ) = @_;
    my $fetch_params = $self->get_fetch_params_sorted($c);
    my $pager_params = $self->get_pager_params($c);

    if ($c->debug) {
        my $w = Catalyst::Utils::term_width() - 20 - 8;
        my $f = Text::SimpleTable->new(20, $w);
        foreach my $k (sort keys %$fetch_params) {
            my $val = ref($fetch_params->{$k}) eq 'ARRAY' ? join(', ', @{ $fetch_params->{$k} }) : $fetch_params->{$k};
            if ((my $new = blessed($val)) && $val->isa("state51::Base")) {
                $new .= ":" . $val->id;
                $val = $new;
            }
            $f->row($k, $val);
        }
        $c->log->debug("fetch_params:\n" . $f->draw);
        my $p = Text::SimpleTable->new(20, $w);
        foreach my $k (sort keys %$pager_params) {
            $p->row($k, $pager_params->{$k});
        }
        $c->log->debug("pager_params:\n" . $p->draw)
    }

    return $self->get_object_pager(
        $fetch_params,$pager_params
    );
}

sub filter : Chained('base_grid')  PathPart('filter') Args(0){
    my ($self, $c) = @_;

    if ($self->set_filter_params($c)) {
        $self->redirect_to_default_page($c);
    }
    else {
        # This is cheating as we break abstraction here.. Sod it :)
        unless (-e $c->path_to('root', $self->action_namespace, 'filter.html')) {
            $c->stash(template => $self->action_namespace . '/default.html');
        }
        $c->detach('default');
    }
}

sub reset_filter : Chained('base_grid')  PathPart('reset_filter') Args(0) ActionClass('REST') {}

sub reset_filter_GET {
    my ($self, $c) = @_;
    $self->reset_filter_params($c);
    $self->redirect_to_default_page($c);
}

sub redirect_to_default_page {
    my ($self, $c) = @_;
    $c->res->redirect($c->uri_for($self->action_for('default'), [ ], $c->req->args->flatten));
    $c->detach;
}

1;

