package CatalystX::state51::ErrorCatcher::RT;
use strict;
use warnings;
use Try::Tiny;
use RT::Client::REST;
use RT::Client::REST::Ticket;
use Data::Dumper;
use DateTime;
use namespace::autoclean;

my $hostname = `hostname --fqdn`;
chomp($hostname);

sub emit {
    my ($self, $c, $output) = @_;
    return if $c->req->user_agent =~ /Test-WWW-Mechanize/;
    return unless $::REPORT_ERRORS_IN_RT;
    local $Data::Dumper::Maxdepth = 2;
    my $config = _check_config($c);
    my $rt = _get_rt_instance($config);
    my $dt = DateTime->now;
    my $appname = ref($c) || $c;
    my $versionstr = ref($c) . '::VERSION';
    my $version = do { no strict 'refs'; $$versionstr };
    my $subject = sprintf('Error executing action: %s on %s: %s Platform: %s',$c->req->action, $appname, $version, $state51::Base::VERSION);
    my $message = sprintf(
        "%s has crashed when a user tried to execute %s (%s) at %s %s \nWeb code version: %s\nPlatform version: %s\nHostname: %s\nUser logged in: %s\nReferer: %s\nURI: %s\n",
        $appname, $c->req->action, $c->req->method, $dt->mdy , $dt->hms, $version, $state51::Base::VERSION,
        $hostname,
        ($c->stash->{role})? $c->stash->{role}->email : '- none -',
        $c->req->referer,
        $c->req->uri,
    );
    my $d = Data::Dumper->new([ $c->session ])->Maxdepth(5);
    my $details = sprintf(
        "%s\n\n*******************\nREQUEST: \n%s \n\n*******************\nSTASH: \n%s \n\n*******************\nSESSION: \n%s \n\n",
        $output, Dumper($c->req), Dumper($c->stash), $d->Dump,
    );

    my $ticket = RT::Client::REST::Ticket->new(rt => $rt);

    my $results;
    my $count = 0;
    try { #Searchs for a open ticket with the same issue searching by subject
        $results = $ticket->search(
            limits => [
                { attribute => 'status', operator => '!=', value => 'resolved' },
                { attribute => 'subject', operator => 'like', value => $subject },
            ],
            orderby => 'subject',
        );
        $count = $results->count;
    };

    if ($count == 0) {
        $ticket = RT::Client::REST::Ticket->new(
            rt => $rt,
            queue => $config->{queue},
            subject => $subject,
            creator => q{CI-Support - Error reporter <ci-support-error@ci-support.com>},
        )->store(text => $message.$details );
    } else {
        my $iterator = $results->get_iterator;
        $ticket = &$iterator;
        $ticket->comment(
            message => $message.$details,
        );
    }
    $c->log->debug('SET last_error_details, last_error_ticket_id:' . $ticket->id) if $c->debug;
    $c->session(last_error_details => $message.$output);
    $c->session(last_error_ticket_id => $ticket->id);
    return;
}

sub _check_config {
    my $c = shift;

    my $config = $c->config->{"CatalystX::state51::ErrorCatcher::RT"};
    if (not defined $config) {
        die __PACKAGE__ . " has no configuration\n";
    }

    if (not defined $config->{server}) {
        die __PACKAGE__ . " has no server\n";
    }

    if (not defined $config->{username}) {
        die __PACKAGE__ . " has no username\n";
    }

    if (not defined $config->{password}) {
        die __PACKAGE__ . " has no password\n";
    }

    if (not defined $config->{queue}) {
        die __PACKAGE__ . " has no queue\n";
    }

    return $config;
}

sub _get_rt_instance {
    my $config = shift;
    my $rt = RT::Client::REST->new(
        server => $config->{server},
        timeout => 10,
    );
    my $user = $config->{username};
    my $pass = $config->{password};

    $rt->login(username => $user, password => $pass);
    return $rt;
}

1;
__END__

